<copyright> TESLA Callable class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Callable.t,v 1.3 1999/09/22 19:54:27 tiggr Exp $</id>
    </copyright>

implementation class
Callable extension tesla: tesla
{
  <doc> Boolean flags to store in the {int} {flags}, so we do not spend a
      whole {boolean} (the size of a byte) on each one.  </doc>
  const CALLABLE_DYNAMIC_IN = 1;
  const CALLABLE_DYNAMIC_OUT = 2;
}

end;

implementation instance
Callable extension tesla
{
  <doc> The formal selector of this callable.  </doc>
  Selector formal_selector;

  <doc> If not {nil}, all local variables and argumenst that are needed in
      the context of directly enclosed blocks - all of them.  </doc>
  MutableIndexed context;

  <doc> The flags.  </doc>
  int flags;
}

<doc> Compute and return the actual selector for an invocation of the
    receiving callable (block or method), when passed the
    {actual_arguments} (expressions) and returning the {actual_return}.
    If {actual_arguments} is {nil}, the selector is computed for the
    formal arguments (and the {CALLABLE_DYNAMIC_IN} and
    {CALLABLE_DYNAMIC_OUT} of the {flags} are set if needed).  </doc>
Selector (result)
  actualSelector (Array, Type) (actual_arguments, actual_return)
{
  MutableByteString td_name_in = [MutableByteString new];
  MutableByteString sel_name = [MutableByteString new];
  Type ar = !actual_return ? [self return_type] : actual_return;
  TypeDescription td_in, td_out;
  boolean dyn_in, dyn_out;

  dyn_out = [ar collectTypesDescription sel_name];
  if (!actual_return && dyn_out)
    flags |= CALLABLE_DYNAMIC_OUT;
  td_out = [TypeDescription withName sel_name];
  // Naughty!
  if ([sel_name length] > 1)
    {
      [sel_name truncate 0];
      [sel_name print ('(', [td_out name], ')')];
    }

  Indexed arguments = [self arguments];
  Indexed name_parts = [self name_parts];
  int i, na = !arguments ? 0 : [arguments length], nm = [name_parts length];
  Indexed aa = !actual_arguments ? arguments : actual_arguments;
  MutableByteString ad = [MutableByteString new];

  for (i = 0; i < nm; i++)
    {
      [sel_name print ('_', ({String s = name_parts[i];}))];
      if (i < na)
	{
	  [ad truncate 0];

	  /* Use the formal type, except if it is the dynamic type, in
             which case we use the actual type (if supplied).  */
	  Type aat = [arguments[i] type];
	  if (aat == dynamic_type)
	    aat = [aa[i] type];
	  dyn_in |= [aat collectTypesDescription ad];

	  [td_name_in print ad];
	  // Naughty!
	  if ([ad length] == 1)
	    [sel_name print ('_', ad)];
	  else
	    [sel_name print ("_(", ad, ')')];
	}
    }
  if (dyn_in && !actual_arguments)
    flags |= CALLABLE_DYNAMIC_IN;

  td_in = [TypeDescription withName td_name_in];
  = [Selector withName sel_name types (td_in, td_out)];
}

<doc> Return some number identifying the receiving callable uniquely between
    its (same class) peers.  </doc>
int (result)
  contextID
{
  <c>
    result = (int) self;
  </c>
}

<doc> Return the {formal_selector}, creating it if necessary.  </doc>
Selector
  formal_selector
{
  if (!formal_selector)
    formal_selector = [self actualSelector (nil, nil)];
  = formal_selector;
}

<doc> Add the variable {v} to the {context}, creating it if necessary.
    </doc>
void
  noteBlockReference BlockDefinition b
		  of VariableSingle v
{
  if (!context)
    context = [MutableObjectArray with v];
  else
    [context add v];
}

void
  precompileDeclarations Cblock parent
		  method Cblock body
{
  if (context != nil)
    {
      [parent addDecl [[CContextDecl alloc] init self]];
      int i, n = [context length];
      for (i = 0; i < n; i++)
	{
	  VariableSingle v = context[i];
	  [v precompileContextInitialization parent method body];
	}
    }
}

<doc> Return the type of value returned by this Callable.  </doc>
deferred Type
  return_type;

OutputStream (of)
  writeContextName OutputStream of
{
  [of print ("_context_", [self contextID])];
}

<doc> Invoke {writeContextName} or print {0}, depending on {context}.  </doc>
TeslaOutput (of)
  writeContextNameOrNil TeslaOutput of
{
  if (!context)
    [of print "0"];
  else
    [self writeContextName of];
}

<doc> If the {context} is not {nil}, emit the declaration of the struct
    containing the context.  </doc>
TeslaOutput (of)
  writeCContextDeclaration TeslaOutput of
{
  if (!context)
    return;
  
  [[[self writeContextName [of print "struct "]] nl] print "{"];
  [of indent 2];

  int i, n = [context length];
  for (i = 0; i < n; i++)
    {
      LocalVariable lv = context[i];
      [[lv writeCContextDeclaration [of nl nil]] print ';'];
    }

  [of indent -2];
  [[[of nl nil] print "};"] nl];
}

end;
