<copyright> TESLA mu.Compound extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998, 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Compound.t,v 1.23 1999/09/22 19:54:27 tiggr Exp $</id>
    </copyright>

implementation class
Compound extension tesla: tesla
{
  <doc> Various flags.  </doc>
  const C_TYPED_EXPRS = 1;
  const LOOP_CONDITION_AFTER = 2;

  <doc> For use within {writeFields}, the indentation of each top
      expression.  </doc>
  static MutableByteString wf_indent;
}

end;

implementation instance
Compound extension tesla
{
  <doc> Any blocks we directly contain.  </doc>
  MutableIndexed blocks;

  <doc> Various flags.  </doc>
  int c_flags;
}

<doc> Pass to the {container}.  </doc>
Callable
  callable
{
  = [Compound (container) callable];
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  int i, n = [expressions length];
  if (n > 0)
    {
      for (i = 0; i < n - 1; i++)
	[expressions[i] coarseUsage 0 parent self body m];
      [expressions[n - 1] coarseUsage f parent self body m];
    }
}

<doc> Compile the {blocks} and instruct our enclosed compounds to do the
    same.  </doc>
TeslaOutput (of)
  compileBlocks TeslaOutput of
{
  int i, n = [expressions length];

  for (i = 0; i < n; i++)
    [expressions[i] compileBlocks of];
}

<doc> Pass to the {container} or issue an error.  </doc>
Loop
  enclosingLoop
{
  if (!container)
    [[[ParseTom error] print "no enclosing loop"] nl];
  else
    = [Compound (container) enclosingLoop];
}

void
  precompileDeclarations Cblock parent
		  method Cblock body
{
  if (variables_ordered != nil)
    {
      int i, n = [variables_ordered length];
      for (i = 0; i < n; i++)
	{
	  LocalVariable lv = variables_ordered[i];
	  [lv precompileCompoundDeclaration parent];
	}
    }

  if (blocks != nil)
    {
      int i, n = [blocks length];
      for (i = 0; i < n; i++)
	{
	  BlockDefinition block = blocks[i];
	  if ([block haveBlockVariables])
	    [parent addDecl [[CBlockDecl alloc] init block]];
	}
    }
}

protected Cexpr (last)
  precompileElements Cblock parent
	      method Cblock body
{
  [self precompileDeclarations parent method body];

  int i, n = [expressions length];
  for (i = 0; i < n; i++)
    last = [expressions[i] precompile parent method body];
}

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  Cblock b = [[Cblock alloc] initWithParent parent];

  if (expr_flags & EXPR_RHS != 0 && type != void_type)
    result = [type temporaryCexpr parent];

  Cexpr last = [self precompileElements b method body];

  if (result != nil)
    [last compile b assignmentTo result];

  [parent add b];
}

<doc> Add the block {b} to the {blocks}, creating the latter if needed.
    </doc>
void
  registerBlock BlockDefinition b
{
  if (!blocks)
    blocks = [MutableObjectArray with b];
  else
    [blocks add b];
}

<doc> If this compound has not been typed yet, first type the elements,
    getting the type from the {last_expression}.  Then pass to {super}.
    </doc>
(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    [self resolveLastExpression (expected, convertible)
	  context context_type nested: (indices, index)];

  if (c_flags & C_TYPED_EXPRS == 0)
    {
      c_flags |= C_TYPED_EXPRS;
      int i, n = [expressions length];
      for (i = 0; i < n; i++)
	{
	  Expression e = expressions[i];
	  Expression f = [e resolve (nil, nil) context context_type];
	  if (e != f)
	    {
	      // No need to check for last_expression changing, since it
	      // was already fully typed.
	      // Thu May 28 13:13:32 1998, tiggr@gerbil.org
	      expressions[i] = f;
	    }
	}
    }

  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

<doc> Resolve the last expression in the receiving compound.  </doc>
Expression
  resolveLastExpression (Cons, Type) (expected, convertible)
		context Meta context_type
		nested: (MutableIntArray, int) (indices, index)
{
  int n = [expressions length];
  if (n > 0)
    {
      Expression e = expressions[n - 1];
      Expression f = [e resolveLastExpression (expected, convertible)
			context context_type nested: (indices, index)];
      if (e != f)
	expressions[n - 1] = last_expression = f;
    }
  if (!last_expression)
    {
      if (flag_warn_empty_compound)
	[[[ParseTom warning self] print "compound without value"] nl];
      last_expression = void_expr;
    }
  type = [last_expression force_type];
  = self;
}

<doc> Emit the expressions.  </doc>
OutputStream
  writeFields OutputStream s
{
  if (!wf_indent)
    wf_indent = [MutableByteString new];

  [wf_indent add ' '];
  [wf_indent add ' '];

  int i, n = [expressions length];
  for (i = 0; i < n; i++)
    s = [[s nl] print (wf_indent, Expression (expressions[i]))];
  [wf_indent truncate [wf_indent length] - 2];
  = s;
}

end;
