<copyright> TESLA mu.Unit extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Unit.t,v 1.26 1998/07/23 15:48:54 tiggr Exp $</id>
    </copyright>

implementation class
ParseUnit extension tesla: tesla

end;

implementation instance
ParseUnit extension tesla

<doc> Be notified that the poser {msub} poses as the posed {msuper}.
    The default implementation does nothing.  </doc>
void
  notePoser Meta msub
      poses Meta msuper
{
  [unit poser msub poses msuper];
}

end;

/******************** mu.Unit (tesla) ********************/

implementation class
mu.Unit extension tesla: tesla
{
  <doc> Boolean flags to store in the {int} {flags}, so we do not spend a
      whole {boolean} (the size of a byte) on each one.  </doc>
  const foo = 1;
}

<doc> Construct a unit containing a single file with the {fname},
    containing classes with the {class_names} and named extensions
    with the {ext_names} (of the form {class_name.extension_name}),
    and depending on the units with the {unit_names}.  Return the
    newly created unit, or {nil} in case of an error.

    The {class_names} are either plain names, declaring the named
    class; or qualified names as in {sub:super}, which declare the
    class {sub} as posing the class {super}.  </doc>
instance (id)
  withSingleFile String fname
	 classes Indexed class_names
      extensions Indexed ext_names
     dependingOn Indexed unit_names
pre
  fname != nil && class_names != nil
{
  mu.Unit unit = [[self alloc] init [File basename fname]
			       directory [File directory-of-file fname]];
  [unit uses [[self builtin_unit] name]];

  TomFile file = [unit fileWithName [[MutableByteString new]
				      print (fname, ".t")]];
  int i, n;

  if (unit_names != nil)
    for ({i = 0; n = [unit_names length];}; i < n; i++)
      [unit uses unit_names[i]];

  for ({i = 0; n = [class_names length];}; i < n; i++)
    {
      String spec = class_names[i], s = spec;
      Indexed a = [s componentsSeparatedBy ':'];
      int alen = [a length];

      if (alen > 2)
	{
	  [[[ParseUnit error] print ("bad posing indication: ", s)] nl];
	  return nil;
	}
      else if (alen > 1)
	s = a[0];

      Meta meta = [Meta withName s file file];
      [file addElement [meta extensionNamed nil]];
      [unit addMeta meta];

      if (alen > 1)
	{
	  s = a[1];
	  Meta posed = [unit metaNamed s recurse: YES];
	  if (!posed)
	    {
	      [[[ParseUnit error] print (spec, ": unknown meta: ", s)] nl];
	      return nil;
	    }

	  [unit poser meta poses posed];
	}
    }

  if (ext_names != nil)
    for ({i = 0; n = [ext_names length];}; i < n; i++)
      {
	String s = ext_names[i];
	Indexed a = [s componentsSeparatedBy '.'];
	if ([a length] != 2)
	  {
	    [[[ParseUnit error] print ("bad extension: ", s)] nl];
	    return nil;
	  }
	String meta_name = a[0];
	Meta meta = [unit metaNamed meta_name];
	if (!meta)
	  {
	    [[[ParseUnit error] print (s, ": unknown meta: ", meta_name)] nl];
	    return nil;
	  }

	String extension_name = a[1];
	[file addElement [Extension ofMeta meta inFile file
			    name: extension_name]];
	[Extension ofMeta [meta meta] inFile file name: extension_name];
      }

  [all_units add unit];
  [all_user_units add unit];

  = unit;
}

end;

implementation instance
mu.Unit extension tesla
{
  <doc> Posing associations.  Each element is a {Cons} dotted pair
      {(sub . super)}.  </doc>
  MutableArray posers;

  <doc> Flags.  </doc>
  int flags;
}

<doc> Tell all {metas} in this unit that {allUnitsLoaded}, and that it is
    time to resolve references, type method invocations, etc.  </doc>
void
  allUnitsLoaded
{
  [[err print ("Preparing unit ", name, "...")] nl];
  Enumerator meta_e = [metas enumerator];
  Instance ins;
  while ({boolean b; (b, ins) = [meta_e next]; b;})
    [ins allUnitsLoaded];
}

<doc> Load all files in this unit.  Then tell all files in this unit
    that their {singleUnitLoaded}.  </doc>
void
  loadFiles
{
  [[err print ("Loading unit ", name, "...")] nl];
  [self setPosers];

  [files makeElementsPerform selector (void loadFile)];

  /* Now everything about this unit is known, check redeclarations and
     such things, before the units using us can confuse things.  */
  [files makeElementsPerform
	 [tom.Invocation of selector (void singleUnitLoaded boolean)
	     with FALSE]];

  /* Pass 2, resolving types.  */
  [files makeElementsPerform
	 [tom.Invocation of selector (void singleUnitLoaded boolean)
	     with TRUE]];

  [self clearPosers];
}

void
  makeMethodsPerform tom.Invocation inv
{
  // We want :)
  // [files do |{ TomFile f | [f perform inv]; }|];
  // Thu Jul 23 13:21:34 1998, tiggr@gerbil.org
  [files makeElementsPerform selector (void makeMethodsPerform tom.Invocation)
	 with inv];
}

<doc> Be notified that the poser {msub} poses as the posed {msuper}.
    </doc>
redefine void
  poser Meta msub
  poses Meta msuper
{
  if (!posers)
    posers = [MutableObjectArray withCapacity 1];
  [posers add [Cons with (msub, msuper)]];
}

<doc> Set up the posing metas of this unit and all units upon which it
    depends.  </doc>
void
  setPosers
{
  [self setPosers ++search_mark];
}

<doc> Hard worker for the above.  </doc>
protected void
  setPosers int k
pre
  mark <= k
post
  mark == k
{
  if (mark == k)
    return;
  mark = k;

  if (used != nil)
    {
      int i, n = [used length];
      for (i = 0; i < n; i++)
	[used[i] setPosers mark];
    }

  if (posers != nil)
    {
      int i, n = [posers length];
      Meta poser, posed;

      for (i = 0; i < n; i++)
	{
	  (poser, posed) = [posers[i] decons];
	  [posed setPoser poser];
	}
    }
}

<doc> Remove any posing by the metas in the receiving unit or any of
    the units upon which it depends.  </doc>
void
  clearPosers
{
  [self clearPosers ++search_mark];
}

<doc> Hard worker for the above.  </doc>
protected void
  clearPosers int k
pre
  mark <= k
post
  mark == k
{
  if (mark == k)
    return;
  mark = k;

  if (used != nil)
    {
      int i, n = [used length];
      for (i = 0; i < n; i++)
	[used[i] clearPosers mark];
    }

  if (posers != nil)
    {
      int i, n = [posers length];
      Meta poser, posed;

      for (i = 0; i < n; i++)
	{
	  (poser, posed) = [posers[i] decons];
	  [posed clearPosing];
	  [poser clearPosing];
	}
    }
}

<doc> Write any files necessary to create or update the output of the
    files in this unit.  If {only_one}, that is the only file to be
    written.  If {single_unit}, the other units are not traversed.  In
    that case, if this is the tom unit also do the {_builtin} unit.
    </doc>
void
  writeCFiles String only_one
pre
  !single_unit || single_unit == self
{
  /* Do the files, or the ONLY_ONE (roaarrrr).  */
  Enumerator files_e = [files enumerator];
  TomFile file;
  while ({boolean b; (b, file) = [files_e next]; b;})
    if (!only_one || [only_one equal [file name]])
      {
	[file writeCFile];
	if (only_one != nil)
	  break;
      }

  if (!only_one)
    {
      [self writeRHFile];
      [self writeRCFile];

      //      if (self == unit_tom && single_unit != nil)
      //	{
      //	  [builtin_unit writeRHFile];
      //	  [builtin_unit writeRCFile];
      //	}
    }
}

<doc> Write the resolved C file.  The precondition states that we
    currently only handle the -1 case.  </doc>
void
  writeRCFile
pre
  single_unit != nil
{
  MutableByteString fn = [[MutableByteString new] print (name, "-r.c")];
  OutputStream of = [File open fn output: YES
			  flags: FILE_NOT_EXIST_CREATE | FILE_EXIST_TRUNCATE];
  of = [[BufferedStream alloc] init of];
  unwind ([[of flushOutput] close])
    ({
      [of print ("/* Your mother was a gorilla.  */
#define TESLA_OUTPUT
#include \"",
		 name, "-r.h\"\n")];

      /* Define type descriptions.  */
      [[TypeDescription used_descriptions] makeElementsPerform
	[tom.Invocation of selector (All writeCDefinition All) with of]];

      /* Define the selectors.  */
      MutableEqSet used_selectors = [Selector used_selectors];
      [used_selectors makeElementsPerform
	[tom.Invocation of selector (All writeCDefinition All) with of]];

      // Shouldn't this paragraph be functionality of the Selector class?
      // Fri Jul 10 11:57:40 1998, tiggr@gerbil.org
      /* Emit the selector table.  */
      [of print "static struct trtd_selectors selectors = {{"];
      int i, n = [used_selectors length];
      [of print (n, ", 0}, {")];
      Enumerator e = [used_selectors enumerator];
      Selector sel;
      while ({boolean b; (b, sel) = [e next]; b;})
	[[sel writeCReferenceName [of print "\n&"]] print ","];
      [of print "\n}};\n"];

      /* Dynamic selector map.  */
      [of print "\nstatic struct trtd_selector_map dynamic_selectors[] = {"];
      MutableEqSet dynamic_formals = [Selector dynamic_formals];
      [dynamic_formals makeElementsPerform
		 [tom.Invocation of selector (All writeCDynamicMap All)
				 with of]];
      [of print "\n};\n"];

      /* Declare the extension descriptions.  */
      MutableArray extensions = [MutableObjectArray new];
      Enumerator files_e = [files enumerator];
      TomFile file;
      while ({boolean b; (b, file) = [files_e next]; b;})
	[file writeRCFile of pass 0 collect extensions];
      int num_extensions = [extensions length];

      /* Metas: define _mr_ and _md_.  */
      MutableArray metas_indexed = [MutableObjectArray new];
      Enumerator metas_e = [metas enumerator];
      Meta meta;
      while ({boolean b; (b, meta) = [metas_e next]; b;})
	[meta writeRCFile of pass 0 collect metas_indexed];
      int num_metas = [metas_indexed length];

      /* Load methods.  */
      Indexed load_methods = [Meta collectLoadMethods [MutableObjectArray new]
				   forUnit self];
      int num_load_methods = [load_methods length];
      if (num_load_methods > 0)
	{
	  for (i = 0; i < num_load_methods; i++)
	    [[load_methods[i] writeCFunctionName [of print "int "]]
	      print " ();\n"];

	  [of print "static struct trtd_load_imp load_imps[] = {"];
	  for (i = 0; i < num_load_methods; i++)
	    {
	      CustomMethod m = load_methods[i];
	      [m writeCFunctionName [of print "{"]];
	      [[[m extension] meta]
		writeCReferenceName [of print ", (void *) &"]];
	      [of print "},\n"];
	    }
	  [of print "};\n"];
	}

      /* Define super references.  */
      EqSet used_super_refs = [SuperReference used_references];
      if (used_super_refs != nil)
	[used_super_refs makeElementsPerform
	    [tom.Invocation of selector (All writeCDefinition All) with of]];

      /* Define the super reference descriptions.  */
      int num_super_refs;
      if (used_super_refs != nil)
	{
	  Enumerator sr_e = [used_super_refs enumerator];
	  SuperReference sr;

	  [of print "\nstatic struct trtd_super super_refs[] = {\n"];
	  while ({boolean b; (b, sr) = [sr_e next]; b;})
	    {
	      [sr writeCInitializer of];
	      num_super_refs++;
	    }
	  [of print "};\n"];
	}

      /* List the extensions.  */
      [[C write [of print "\nstatic struct trtd_extension *trtd_extensions_"]
	  CName name] print "[] = {\n"];
      [Meta writeCExtensionList of forUnit self];
      [of print "};\n"];

      /* List the metas (class objects only).  */
      [[C write [of print "\nstatic struct trt_class *trtd_metas_"]
	  CName name] print "[] = {\n"];
      for (i = 0; i < num_metas; i++)
	{
	  Meta m = metas_indexed[i];
	  [[m writeCMetaDefinitionName [of print "(void *) &"]] print ",\n"];
	}
      [of print "};\n"];

      /* Describe this unit.  */
      [C write [of print "\nstatic struct trtd_unit _ud_"] CName name];
      [of print (" = {\n{\"", name, "\", ", [name length], "},\n")];
      [[C write [of print "trtd_extensions_"] CName name]
	print (", ", num_extensions, ",\n")];
      [[C write [of print "trtd_metas_"] CName name]
	print (", ", num_metas, "\n};\n")];

      /* List the units.  */
      [of print "\nstatic struct trtd_unit *trtd_units[] = {\n"];
      [[C write [of print "&_ud_"] CName name] print "\n};\n"];

      /* Register the posers.  */
      int num_posers;
      if (posers != nil)
	{
	  int i;
	  num_posers = [posers length];
	  [of print "\nstatic struct trtd_poser posers[] = {\n"];
	  for (i = 0; i < num_posers; i++)
	    {
	      Meta msub, msuper;
	      (msub, msuper) = [posers[i] decons];
	      [msub writeCInitializer of posing msuper];
	    }
	  [of print "};\n"];
	  /* Count instances and classes separately.  */
	  num_posers *= 2;
	}

      /* Module information.  */
      CustomMethod main_method = [CustomMethod main];
      [C write [of print "\nstruct trtd_module trt_module_info_"] CName name];
      {
	[of print " = {\ntrtd_units, 1, &selectors,\n"];

	/* Dynamic selectors.  */
	if (!dynamic_formals)
	  [of print "0, 0,\n"];
	else
	  [of print ("dynamic_selectors, ", [Selector num_dynamic_map_entries],
		     ",\n")];

	/* Super references.  */
	if (!num_super_refs)
	  [of print "0, 0,\n"];
	else
	  [of print ("super_refs, ", num_super_refs, ",\n")];

	/* Load imps.  */
	if (num_load_methods)
	  [of print ("load_imps, ", num_load_methods, ",\n")];
	else
	  [of print "0, 0,\n"];

	/* Main method.  */
	if (!main_method)
	  [of print "0, 0,\n"];
	else
	  [[[main_method meta]
	     writeCReferenceName [[main_selector writeCReferenceName
						 [of print "&"]]
				   print ", (void *) &"]] print ",\n"];

	/* Posers.  */
	if (num_posers > 0)
	  [of print ("posers, ", num_posers, "\n")];
	else
	  [of print "0, 0\n"];
      }
      [of print "};\n"];

      /* Unit registration function.  */
      [[C write [of print "\nvoid trt_register_unit_"] CName name]
	print " (void) __attribute__ ((constructor));\n"];
      [[C write [of print "void\ntrt_register_unit_"] CName name]
	print " (void)\n{\n  static char initialized = 0;\n  if (initialized)\n    return;\n  initialized = 1;\n"];
      if (used != nil)
	{
	  int i, n = [used length];
	  for (i = 0; i < n; i++)
	    {
	      mu.Unit u = used[i];
	      if (u != builtin_unit)
		[[C write [of print "  trt_register_unit_"] CName [u name]]
		  print " ();\n"];
	    }
	}
      [C write [of print "  trt_resolve_module (&trt_module_info_"] CName name];
      [of print ");\n}\n"];

      /* Set trt_main_registration.  */
      if (main_method != nil)
	{
	  [of print "void (*trt_main_registration)(void) = trt_register_unit_"];
	  [[C write of CName name] print ";\n"];
	}
    });
}

<doc> Write the resolved header file.  The precondition states that we
    currently only handle the -1 case.  </doc>
void
  writeRHFile
pre
  single_unit != nil
{
  MutableByteString fn = [[MutableByteString new] print (name, "-r.h")];
  OutputStream of = [File open fn output: YES
			  flags: FILE_NOT_EXIST_CREATE | FILE_EXIST_TRUNCATE];
  of = [[BufferedStream alloc] init of];
  unwind ([[of flushOutput] close])
    ({
      [of print "/* Your mother was a gorilla.  */\n#ifndef __include_"];
      [[C write of CName name] print "_resolved\n#define __include_"];
      [[C write of CName name]
	print "_resolved\n#ifndef CURRENT_UNIT\n#define CURRENT_UNIT "];
      [C write of CName name];

      if (self == builtin_unit)
	{
	  /* If we are the top unit, we won't get tom/tom-r.h, and we
             must declare the State meta-class object.  */
	  [of print "\nextern struct trt_class _md_m_tom_State;"];
	}

      [of print "\n#endif\n#include <tom/trt.h>\n"];

      /* Include other units' RH files.  */
      if (used != nil)
	{
	  int i, n = [used length];
	  for (i = 0; i < n; i++)
	    {
	      mu.Unit u = used[i];
	      String s = [used[i] name];

	      if (u != builtin_unit || self == unit_tom)
		[of print ("#include <", s, "/", s, "-r.h>\n")];
	    }
	}

      [[TypeDescription used_descriptions] makeElementsPerform
	[tom.Invocation of selector (All writeCDeclaration All) with of]];
      [[Selector used_selectors] makeElementsPerform
	[tom.Invocation of selector (All writeCDeclaration All) with of]];

      Keyed sr = [SuperReference used_references];
      if (sr != nil)
	[sr makeElementsPerform
	    [tom.Invocation of selector (All writeCDeclaration All)
		with of]];

      int pass;
      for (pass = 0; pass < 2; pass++)
	{
	  Enumerator metas_e = [metas enumerator];
	  Meta meta;
	  while ({boolean b; (b, meta) = [metas_e next]; b;})
	    [meta writeRHFile of pass pass];

	  Enumerator files_e = [files enumerator];
	  TomFile file;
	  while ({boolean b; (b, file) = [files_e next]; b;})
	    [file writeRHFile of pass pass];
	}

      [of print "#endif\n"];
    });
}

end; 
