<copyright> TIG tig class.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1999 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: tig.t,v 1.2 1999/10/12 21:24:32 tiggr Exp $</id>
    </copyright>

implementation class
tig: stdio, mu, tom.Constants
{
  <doc> Whether all output files are to be written, irrespective of there
      existence and not being out-of-date.  </doc>
  static boolean flag_force_output;

  <doc> How much do we say.  </doc>
  static int flag_verbosity;

  <doc> The unit of which we are generating the interfaces.  </doc>
  static public mu.Unit top_unit;
}

<doc> Initialize.  </doc>
void
  initialize
{
  [mu init TRUE];
}

<doc> Get the ball rolling, and afterwards report the errors.  </doc>
int (rc)
  main Array arguments
{
  = [self real_main arguments];
  [self report [ParseUnit self] errors "unit"];
  [self report [ParseTom self] errors "tom"];
}

<doc> The real thing.  </doc>
int
  real_main Array arguments
{
  int i, n, arg, argc = [arguments length];
  MutableObjectArray uses_units;
  String unit_name;

  err = [BufferedStream with err];

  for (arg = 0; arg < argc; arg++)
    {
      String s = arguments[arg];
      if (s[0] != '-')
	break;
      if ([s equal "--version"])
	{
	  if (flag_verbosity > 0)
	    [[out print ("tig ", TIG_VERSION)] nl];
	  else
	    [[out print TIG_VERSION] nl];
	  return 0;
	}
      else if ([s equal "-F"])
	flag_force_output = TRUE;
      else if ([s equal "-v"])
	flag_verbosity++;
      else if (arg < argc - 1)
	{
	  String next_arg = arguments[++arg];

	  if ([s equal "-I"])
	    [Parser addToPath next_arg];
	  else if ([s equal "-u"])
	    unit_name = next_arg;
	  else
	    [self usage s];
	}
      else
	[self usage s];
    }

  if (!unit_name)
    [self usage "missing -u"];

  [self initialize];

  /* Load the .u files.  */
  top_unit = [mu.Unit named unit_name load: YES];
  if (!top_unit || [ParseUnit num_errors] > 0)
    return 1;

  [top_unit setPosers];
  [top_unit loadFiles];

  = [ParseTom num_errors];
}

<doc> Report the number of errors and warnings for the parser class {pc}.
    </doc>
void
  report class (Parser) pc
  errors String d
{
  int nw = [pc num_warnings];
  int ne = [pc num_errors];

  if (nw != 0 || ne != 0)
    [[[pc report] print (d, ": ", ne, " error", ne == 1 ? "" : "s",
			 ", ", nw, " warning", nw == 1 ? "" : "s")] nl];
}

<doc> Output usage information and exit.  </doc>
void
  usage String opt_arg
{
  if (opt_arg != nil)
    [[err print ("bad option or argument: ", opt_arg)] nl];
  [[err print ("usage: ", [Runtime program_name],
	       " -u unit [options]
regular options:
 -I dir   add DIR to the include path
 -F       write all files that can be written, acting as if they are absent
default options:
 --version     report the version of tig and exit
 -v            be verbose")] nl];
  [Runtime exit 2];
}

end;

implementation instance tig end; 
